unit Main;

interface

uses
  ShLwApi, Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, XPMan, ExtCtrls, ImgList, StdCtrls, Buttons, ComCtrls, Registry,
  Menus;

type
  TMainForm = class(TForm)
    XPManifest1: TXPManifest;
    PreviewPanel: TPanel;
    Preview: TImage;
    SaveBtn: TBitBtn;
    ColorDialog: TColorDialog;
    ColorShape: TShape;
    Label2: TLabel;
    InvertColors: TCheckBox;
    Method: TTrackBar;
    Transparent: TCheckBox;
    Label1: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Visibility: TTrackBar;
    Label5: TLabel;
    Label6: TLabel;
    Label7: TLabel;
    Saturation: TTrackBar;
    Label8: TLabel;
    Label9: TLabel;
    Label10: TLabel;
    ResetBtn: TBitBtn;
    UITimer: TTimer;
    SaveDialog: TSaveDialog;
    AboutBtn: TBitBtn;
    Presets: TBitBtn;
    PresetMenu: TPopupMenu;
    Standard1: TMenuItem;
    Blue1: TMenuItem;
    Green1: TMenuItem;
    Purple1: TMenuItem;
    Red1: TMenuItem;
    Yellow1: TMenuItem;
    N1: TMenuItem;
    White1: TMenuItem;
    Black1: TMenuItem;
    Glassy1: TMenuItem;
    Hardwhite1: TMenuItem;
    N2: TMenuItem;
    N3: TMenuItem;
    Darkblack1: TMenuItem;
    Hue: TTrackBar;
    Label11: TLabel;
    Label12: TLabel;
    HueShape: TShape;
    Neon1: TMenuItem;
    Glassy2: TMenuItem;
    Neon2: TMenuItem;
    MenuUserPresets: TMenuItem;
    MenuUserPreset: TMenuItem;
    N4: TMenuItem;
    MenuSaveUserPreset: TMenuItem;
    procedure FormDestroy(Sender: TObject);
    procedure MenuSaveUserPresetClick(Sender: TObject);
    procedure MenuUserPresetClick(Sender: TObject);
    procedure HueShapeMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PresetMenuClick(Sender: TObject);
    procedure PresetsClick(Sender: TObject);
    procedure AboutBtnClick(Sender: TObject);
    procedure SaveBtnClick(Sender: TObject);
    procedure UITimerTimer(Sender: TObject);
    procedure ResetBtnClick(Sender: TObject);
    procedure RedrawEvent(Sender: TObject);
    procedure ColorShapeMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormCreate(Sender: TObject);
  private
    procedure UpdatePreview;
    function Colorize(C: TColor): TColor;
    procedure LoadPreset(FN: string);
    procedure SavePreset(FN: string);
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.dfm}

var
  //ColorToIndex: array[0..$FFFFFF] of Byte;
  //IndexToColor: array[Byte, 0..2] of Cardinal;
  //ColorNr: Integer;
  IsWaterColor: array[0..$FFFFFF] of Boolean;
  Data: array[0..1] of record
    BinaryData: string;
    Patches: array[1..5000] of record
      Address: Integer;
      Color: Integer;
      end;
    PatchNr: Integer;
    Bitmap: TMemoryStream;
    end;

function Swap(C: TColor): TColor;
begin
  Result:=RGB(GetBValue(C),GetGValue(C),GetRValue(C));
end;

procedure LoadFile(Name: string; Index: Integer);
var
  T: Text;
  F: file;
  B: TFileStream;
  S: string;
  Adr1, Adr2, Adr3: Integer;
begin
  with Data[Index] do
    begin
    Assign(F, Name+'.dir');
    Reset(F, 1);
    SetLength(BinaryData, FileSize(F));
    BlockRead(F, BinaryData[1], Length(BinaryData));
    Close(F);

    Assign(T, Name+'.txt');
    Reset(T);
    ReadLn(T);
    Adr1:=-1;Adr2:=-1;
    while not EoF(T) do
      begin
      ReadLn(T, S);
      Adr3:=Adr2;Adr2:=Adr1;
      Adr1:=StrToInt('$'+Copy(S, 1, 8));
      if not(StrToInt('$'+Copy(S, 14, 2))-StrToInt('$'+Copy(S, 11, 2)) in [1..3]) then
        Adr1:=-1;

      if Adr1=Adr3+2 then
        begin
        Inc(PatchNr);
        with Patches[PatchNr] do
          begin
          Address:=Adr3;
          Color:=0;
          Move(BinaryData[Address+1], Color, 3);
          Color:=Swap(Color);
          IsWaterColor[Color]:=True;
          end;
        Adr1:=-1;Adr2:=-1;
        end
      else
        Adr1:=Adr1+1-1;
      end;
    Close(T);

    B:=TFileStream.Create(Name+'.bmp', fmOpenRead);
    Bitmap:=TMemoryStream.Create;
    Bitmap.CopyFrom(B, B.Size);
    B.Free;
    end;
end;

procedure TMainForm.ColorShapeMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  ColorDialog.Color:=ColorShape.Brush.Color;
  if ColorDialog.Execute then
   if (ColorDialog.Color<0)or(ColorDialog.Color>$FFFFFF) then
    ColorShape.Brush.Color:=0
   else
    ColorShape.Brush.Color:=ColorDialog.Color;
  UpdatePreview;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  LoadFile('Solid', 0);
  LoadFile('Transparent', 1);
  UpdatePreview;
  with TRegistry.Create do
    try
      RootKey := HKEY_CURRENT_USER;
      OpenKey('Software\Team17SoftwareLTD\WormsArmageddon', False);
      if ReadString('PATH')<>'' then
       if DirectoryExists(ReadString('PATH')+'\DATA\Water\Blue') then
        SaveDialog.InitialDir:=ReadString('PATH')+'\DATA\Water\Blue'
       else
       if DirectoryExists(ReadString('PATH')+'\DATA') then
        SaveDialog.InitialDir:=ReadString('PATH')+'\DATA';
    finally
      Free;
      end;
  if not DirectoryExists('Presets') then
    MkDir('Presets');
  try
    LoadPreset('Presets\Last Configuration.txt');
  except
    ResetBtnClick(nil);
    end;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  SavePreset('Presets\Last Configuration.txt');
end;

procedure TMainForm.HueShapeMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  ColorDialog.Color:=HueShape.Brush.Color;
  if ColorDialog.Execute then
    HueShape.Brush.Color:=ColorDialog.Color;
  UpdatePreview;
end;

var
  PresetName: string = '';

procedure TMainForm.MenuSaveUserPresetClick(Sender: TObject);
begin
  if InputQuery('Enter a preset name', 'Preset name:', PresetName) then
   if (not FileExists('Presets\'+PresetName+'.txt'))or(MessageBox(Handle, 'A preset with such a name already exists, overwrite?', 'Question', MB_YESNO or MB_ICONQUESTION)=IDYES) then
    begin
    SavePreset('Presets\'+PresetName+'.txt');
    end;
end;

procedure TMainForm.MenuUserPresetClick(Sender: TObject);
begin
  LoadPreset('Presets\' + TMenuItem(Sender).Hint+ '.txt');
  UpdatePreview;
end;

procedure TMainForm.PresetsClick(Sender: TObject);
var
  P: TPoint;
  M: TMenuItem;
  F: TSearchRec;
begin
  P:=Presets.ClientToScreen(Point(0,Presets.Height));
  MenuUserPresets.Clear;
  if FindFirst('Presets\*.txt', faArchive, F)=0 then
    begin
    repeat
      if LowerCase(Copy(F.Name, Length(F.Name)-3, 4))='.txt' then
        begin
        M:=TMenuItem.Create(MenuUserPresets);
        M.Caption:=Copy(F.Name, 1, Length(F.Name)-4);
        M.Hint:=Copy(F.Name, 1, Length(F.Name)-4);
        M.OnClick:=MenuUserPresetClick;
        MenuUserPresets.Add(M);
        end;
    until FindNext(F)<>0;
    FindClose(F);

    M:=TMenuItem.Create(MenuUserPresets);
    M.Caption:='-';
    MenuUserPresets.Add(M);
    end;
  M:=TMenuItem.Create(MenuUserPresets);
  M.Caption:='Save preset...';
  M.OnClick:=MenuSaveUserPresetClick;
  MenuUserPresets.Add(M);

  PresetMenu.Popup(P.X,P.Y);
end;

procedure TMainForm.RedrawEvent(Sender: TObject);
begin
  UpdatePreview;
end;

function Ipol(r:real;b1,b2:Integer):Integer;
begin
  //Result:=b1+Round((integer(b2)-b1)*r);
  Result:=Round(B1*(1-R) + B2*R); 
  {if b1<b2 then
    begin
    if Result<b1 then Result:=b1;
    if Result>b2 then Result:=b2;
    end
  else
    begin
    if Result>b1 then Result:=b1;
    if Result<b2 then Result:=b2;
    end}
  if Result>255 then Result:=255;
  if Result<0 then Result:=0;
end;

function ColIpol(r:real;c1,c2:TColor):TColor;
begin
  Result:=RGB(Ipol(r,GetRValue(c1),GetRValue(c2)),
              Ipol(r,GetGValue(c1),GetGValue(c2)),
              Ipol(r,GetBValue(c1),GetBValue(c2)));
end;

const
  Base : Integer = $323B7E;

function BoundColor(C: TColor): TColor;
begin
  Result:=C;
  if Result>255 then Result:=255;
  if Result<0 then Result:=0;
end;

var
  Direction: Integer;

function F1(C, Base, NewBase: Integer): Integer;
begin
  Result:=(C - Base)*Direction + NewBase;
  if Result>255 then Result:=255;
  if Result<0 then Result:=0;
end;

function F2(C, Base, NewBase: Integer): Integer;
begin
  if Direction>0 then
    if Base=0 then
      Result:=0
    else
      Result:=C * NewBase div Base
  else
    if NewBase=0 then
      Result:=0
    else
      Result:=C * Base div NewBase;
  if Result>255 then Result:=255;
  if Result<0 then Result:=0;
end;

procedure TMainForm.ResetBtnClick(Sender: TObject);
begin
  ColorShape.Brush.Color:=$007E3B32;
  Method.Position:=0;
  Visibility.Position:=100;
  Saturation.Position:=100;
  Hue.Position:=0;
  HueShape.Brush.Color:=$007E3B32;
  Transparent.Checked:=False;
  InvertColors.Checked:=False;
  UpdatePreview;
end;

function SetHueSat(C: TColor; HueSatColor: TColor): TColor;
var
  MH, ML, MS, CH, CL, CS: Word;
begin
  ColorRGBToHLS(HueSatColor, MH, ML, MS);
  ColorRGBToHLS(C, CH, CL, CS);
  Result:=ColorHLSToRGB(MH, CL, MS);
end;

function TMainForm.Colorize(C: TColor): TColor;
var
  M1, M2: TColor;
  Z: Integer;
  MyBase, NewBase: Integer;
begin
  if InvertColors.Checked then
    Direction:=-1
  else
    Direction:=1;

  NewBase:=Swap(ColorShape.Brush.Color);

  // saturation
  Z:=(GetRValue(Base)+GetGValue(Base)+GetBValue(Base)) div 3;
  MyBase:=ColIPol(Saturation.Position/100, RGB(Z, Z, Z), Base);

  Z:=(GetRValue(C)+GetGValue(C)+GetBValue(C)) div 3;
  C:=ColIPol(Saturation.Position/100, RGB(Z, Z, Z), C);

  // method - recoloration takes place here
  if C=0 then
    M1:=0
  else
    M1:=RGB( F1(GetRValue(C), GetRValue(MyBase), GetRValue(NewBase)),
             F1(GetGValue(C), GetGValue(MyBase), GetGValue(NewBase)),
             F1(GetBValue(C), GetBValue(MyBase), GetBValue(NewBase)));
  M2:=RGB( F2(GetRValue(C), GetRValue(MyBase), GetRValue(NewBase)),
           F2(GetGValue(C), GetGValue(MyBase), GetGValue(NewBase)),
           F2(GetBValue(C), GetBValue(MyBase), GetBValue(NewBase)));
  C:=ColIPol(Method.Position/100, M1, M2);

  // visibility - sort of alpha-blend with new water color, can be > 100
  C:=ColIPol(Visibility.Position/100, NewBase, C);

  // tint - recoloring of objects
  C:=ColIPol(Hue.Position/100, C, SetHueSat(C, Swap(HueShape.Brush.Color)));

  Result:=C;
end;

procedure TMainForm.LoadPreset(FN: string);
var
  F: TextFile;
  function ReadInt : Integer; begin ReadLn(F, Result) end;
  function ReadBool: Boolean; var S: string; begin ReadLn(F, S); Result:=S='TRUE' end;
begin
  AssignFile(F, FN);
  Reset(F);
  ColorShape.Brush.Color := ReadInt;
  Method.Position := ReadInt;
  Visibility.Position := ReadInt;
  Saturation.Position := ReadInt;
  Hue.Position := ReadInt;
  HueShape.Brush.Color := ReadInt;
  Transparent.Checked := ReadBool;
  InvertColors.Checked := ReadBool;
  CloseFile(F);
end;

procedure TMainForm.SavePreset(FN: string);
var
  F: TextFile;
begin
  AssignFile(F, FN);
  ReWrite(F);
  WriteLn(F, ColorShape.Brush.Color);
  WriteLn(F, Method.Position);
  WriteLn(F, Visibility.Position);
  WriteLn(F, Saturation.Position);
  WriteLn(F, Hue.Position);
  WriteLn(F, HueShape.Brush.Color);
  WriteLn(F, Transparent.Checked);
  WriteLn(F, InvertColors.Checked);
  CloseFile(F);
end;

procedure TMainForm.UITimerTimer(Sender: TObject);
begin
  ClientHeight:=SaveBtn.Top+SaveBtn.Height+SaveBtn.Left;
end;

procedure TMainForm.UpdatePreview;
var
  M: TMemoryStream;
  I: Integer;
  Palette: array[byte]of TColor;
begin
  HueShape.Visible:=Hue.Position > 0;
  with Data[Ord(Transparent.Checked)] do
    begin
    M:=TMemoryStream.Create;
    Bitmap.Position:=0;
    M.CopyFrom(Bitmap, Bitmap.Size);
    M.Position:=SizeOf(TBitmapFileHeader)+SizeOf(TBitmapInfoHeader);
    M.Read(Palette, SizeOf(Palette));
    for I:=1 to 255 do
     if IsWaterColor[Palette[I]] then
      Palette[I]:=Colorize(Palette[I]);
    M.Position:=SizeOf(TBitmapFileHeader)+SizeOf(TBitmapInfoHeader);
    M.Write(Palette, SizeOf(Palette));
    M.Position:=0;
    Preview.Picture.Bitmap.LoadFromStream(M);
    M.Free;
    end;
end;

procedure TMainForm.SaveBtnClick(Sender: TObject);
var
  S: string;
  I, C: Integer;
  F: File;
begin
  SaveDialog.FileName:='Water.dir';
  if SaveDialog.Execute then
    with Data[Ord(Transparent.Checked)] do
      begin
      if Transparent.Checked then
        MessageBox(Handle,
          'Warning: transparent water can give you an'#13#10+
          'unfair advantage, and can be compared to cheating.'#13#10+
          'Please use this feature with responsibility.', 'Transparent water', MB_ICONWARNING);

      S:=BinaryData;
      for I:=1 to PatchNr do
       with Patches[I] do
        begin
        C:=Colorize(Color);
        C:=Swap(C);
        Move(C, S[Address+1], 3);
        end;
      AssignFile(F, SaveDialog.FileName);
      ReWrite(F, 1);
      BlockWrite(F, S[1], Length(S));
      CloseFile(F);
      SaveDialog.InitialDir:=ExtractFileDir(SaveDialog.FileName);
      end;
end;

procedure TMainForm.AboutBtnClick(Sender: TObject);
begin
  MessageBox(Handle,
    'Advanced Worms Water Editor, v1.1'#13#10+
    'by The_CyberShadow'#13#10+
    '<thecybershadow@gmail.com>'#13#10+
    ''#13#10+
    'Thanks to bonz and NotWorthy for the ideas :)'#13#10+
    ''#13#10+
    'Program homepage: http://worms.thecybershadow.net/misc/'#13#10+
    ''#13#10+
    'This was mostly coded in one long, sleepless night ;)'#13#10,
    'About this program', MB_ICONINFORMATION);
end;

procedure TMainForm.PresetMenuClick(Sender: TObject);
begin
  case TComponent(Sender).Tag of
    1:begin       // Blue
      ColorShape.Brush.Color:=$7E3B32;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=False;
      InvertColors.Checked:=False;
      end;
    2:begin       // Green
      ColorShape.Brush.Color:=$7E8032;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=False;
      InvertColors.Checked:=False;
      end;
    3:begin       // Purple
      ColorShape.Brush.Color:=$7E3470;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=False;
      InvertColors.Checked:=False;
      end;
    4:begin       // Red
      ColorShape.Brush.Color:=$323B80;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=False;
      InvertColors.Checked:=False;
      end;
    5:begin       // Yellow
      ColorShape.Brush.Color:=$326B80;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=False;
      InvertColors.Checked:=False;
      end;

   11:begin       // White
      ColorShape.Brush.Color:=$FFFFFF;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=50;
      Hue.Position:=100;
      HueShape.Brush.Color:=$FFFFFF;
      Transparent.Checked:=False;
      InvertColors.Checked:=True;
      end;
   12:begin       // Hard White
      ColorShape.Brush.Color:=$FFFFFF;
      Method.Position:=0;
      Visibility.Position:=200;
      Saturation.Position:=50;
      Hue.Position:=100;
      HueShape.Brush.Color:=$FFFFFF;
      Transparent.Checked:=True;
      InvertColors.Checked:=True;
      end;

   21:begin       // Black
      ColorShape.Brush.Color:=$000000;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=True;
      InvertColors.Checked:=False;
      end;
   22:begin       // Dark Black
      ColorShape.Brush.Color:=$000000;
      Method.Position:=0;
      Visibility.Position:=50;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=False;
      InvertColors.Checked:=False;
      end;
   23:begin       // Neon
      ColorShape.Brush.Color:=$000000;
      Method.Position:=0;
      Visibility.Position:=200;
      Saturation.Position:=200;
      Hue.Position:=100;
      HueShape.Brush.Color:=clLime;
      Transparent.Checked:=True;
      InvertColors.Checked:=False;
      end;
   24:begin       // Tar
      ColorShape.Brush.Color:=$000000;
      Method.Position:=0;
      Visibility.Position:=125;
      Saturation.Position:=200;
      Hue.Position:=0;
      Transparent.Checked:=False;
      InvertColors.Checked:=False;
      end;

   31:begin       // Glassy
      ColorShape.Brush.Color:=$806000;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=0;
      Transparent.Checked:=True;
      InvertColors.Checked:=False;
      end;
   32:begin       // Lava
      ColorShape.Brush.Color:=$0000FF;
      Method.Position:=0;
      Visibility.Position:=100;
      Saturation.Position:=100;
      Hue.Position:=100;
      HueShape.Brush.Color:=$0000FF;
      Transparent.Checked:=False;
      InvertColors.Checked:=True;
      end;
    end;
  UpdatePreview;
end;

end.
